## NOTE: THIS FILE IS GENERATED BY EPSCRIPT! DO NOT MODITY
from eudplib import *
from eudplib.core.eudfunc import EUDTraceLog, EUDTracedFunc, EUDTracedTypedFunc, EUDTracedMethod, EUDTracedTypedMethod
from eudplib.epscript.helper import _RELIMP, _TYGV, _TYSV, _TYLV, _CGFW, _ARR, _VARR, _SRET, _SV, _ATTW, _ARRW, _ATTC, _ARRC, _L2V, _LSH, _ALL
# (Line 1) function test_attribute() {
@EUDFunc
def f_test_attribute():
    # (Line 2) var ret = 0;
    ret = _TYLV([None], [0])
    # (Line 3) const queue = EUDQueue(10)();
    queue = EUDQueue(10)()
    # (Line 4) if (queue.length == 0) ret += 1 << 0;
    if EUDIf()(_ATTC(queue, 'length') == 0):
        ret.__iadd__(_LSH(1,0))
        # (Line 5) queue.append(0);
    EUDEndIf()
    queue.append(0)
    # (Line 6) if (queue.length == 1) ret += 1 << 1;
    if EUDIf()(_ATTC(queue, 'length') == 1):
        ret.__iadd__(_LSH(1,1))
        # (Line 8) const deque = EUDDeque(10)();
    EUDEndIf()
    deque = EUDDeque(10)()
    # (Line 9) if (deque.length == 0) ret += 1 << 2;
    if EUDIf()(_ATTC(deque, 'length') == 0):
        ret.__iadd__(_LSH(1,2))
        # (Line 10) deque.append(0);
    EUDEndIf()
    deque.append(0)
    # (Line 11) if (deque.length == 1) ret += 1 << 3;
    if EUDIf()(_ATTC(deque, 'length') == 1):
        ret.__iadd__(_LSH(1,3))
        # (Line 12) return ret;
    EUDEndIf()
    EUDReturn(ret)
    # (Line 13) }



## NOTE: THIS FILE IS GENERATED BY EPSCRIPT! DO NOT MODITY
from eudplib import *
from eudplib.core.eudfunc import EUDTraceLog, EUDTracedFunc, EUDTracedTypedFunc, EUDTracedMethod, EUDTracedTypedMethod
from eudplib.epscript.helper import _RELIMP, _TYGV, _TYSV, _TYLV, _CGFW, _ARR, _VARR, _SRET, _SV, _ATTW, _ARRW, _ATTC, _ARRC, _L2V, _LSH, _ALL
# (Line 1) function test(fn: EUDFuncPtr(0, 0)) {
@EUDTypedFunc([EUDFuncPtr(0, 0)])
def f_test(fn):
    # (Line 2) fn();
    fn()
    # (Line 3) }
    # (Line 5) EUDOnStart(function () { test(function () {}); });

@EUDFunc
def _lambda1():
    @EUDFunc
    def _lambda2():
        pass

    f_test(_lambda2)

EUDOnStart(_lambda1)



# Copyright 2014 by trgk.
# All rights reserved.
# This file is part of EUD python library (eudplib),
# and is released under "MIT License Agreement". Please see the LICENSE
# file that should have been included as part of this package.

from ...utils import EPD, unProxy
from .. import allocator as ac
from .. import rawtrigger as rt
from .. import variable as ev
from ..eudfunc.eudf import EUDFunc, _EUDPredefineParam, _EUDPredefineReturn
from ..variable import SeqCompute
from ..variable.evcommon import _selfadder, _xv
from .muldiv import _quot

_x = _xv[0]


def f_bitand(a, b):
    """Calculate a & b"""
    return a & b


def f_bitor(a, b):
    """Calculate a | b"""
    return a | b


def f_bitxor(a, b):
    """Calculate a ^ b"""
    return a ^ b


def f_bitnand(a, b):
    """Calculate ~(a & b)"""
    try:
        return (a & b).iinvert()
    except AttributeError:
        return ~(a & b)


def f_bitnor(a, b):
    """Calculate ~(a | b)"""
    try:
        return (a | b).iinvert()
    except AttributeError:
        return ~(a | b)


def f_bitnxor(a, b):
    """Calculate ~(a ^ b)"""
    return f_bitxor(a, ~b)


def f_bitnot(a):
    """Calculate ~a"""
    return ~a


# -------


@EUDFunc
def f_bitsplit(a):
    """Splits bit of given number

    :returns: int bits[32];  // bits[i] = (ith bit from LSB of a is set)
    """
    bits = ev.EUDCreateVariables(32)
    rt.RawTrigger(actions=[bits[i].SetNumber(0) for i in range(32)])
    for i in range(31, -1, -1):
        rt.RawTrigger(conditions=a.AtLeastX(1, 2**i), actions=bits[i].SetNumber(1))
    return bits


# -------


@_EUDPredefineParam(1)
@_EUDPredefineReturn(1, 2)
@EUDFunc
def _exp2_vv(n):
    ret = _exp2_vv._frets[0]
    ret << 0
    for i in range(32):
        rt.RawTrigger(conditions=[n == i], actions=ret.SetNumber(2**i))
    # return ret


def _exp2(n):
    if isinstance(n, int):
        if n >= 32:
            return 0
        return 1 << n

    return _exp2_vv(n)


@EUDFunc
def _f_bitrshift(a, b):
    ret = ev.EUDVariable()

    skip_check = ac.Forward()
    exponential = ac.Forward()
    skip_to_here = ac.Forward()
    return_function = ac.Forward()

    skip_check << rt.RawTrigger(
        nextptr=exponential,
        conditions=b.AtLeast(32),
        actions=[ret.SetNumber(0), rt.SetNextPtr(skip_check, skip_to_here)],
    )
    skip_to_here << rt.RawTrigger(
        nextptr=return_function, actions=rt.SetNextPtr(skip_check, exponential)
    )
    exponential << rt.NextTrigger()
    ret << _quot(a, _exp2_vv(b))
    return_function << rt.NextTrigger()
    return ret


@EUDFunc
def _f_bitlshift(a, b):
    loopstart = ac.Forward()
    loopend = ac.Forward()
    loopcnt = ac.Forward()

    rt.RawTrigger(actions=[rt.SetNextPtr(a.GetVTable(), loopcnt), *a.QueueAddTo(a)])

    loopstart << rt.RawTrigger(
        nextptr=loopend,
        conditions=[a.AtLeast(1), b.AtLeast(1)],
        actions=rt.SetNextPtr(loopstart, a.GetVTable()),
    )
    loopcnt << rt.RawTrigger(
        nextptr=loopstart,
        actions=[b.SubtractNumber(1), rt.SetNextPtr(loopstart, loopend)],
    )
    loopend << rt.NextTrigger()

    return a


def f_bitlshift(a, b, _fdict={}, **kwargs):
    """Calculate a << b"""
    a, b = unProxy(a), unProxy(b)
    if not isinstance(a, ev.EUDVariable) and not isinstance(b, ev.EUDVariable):
        if "ret" in kwargs:
            SeqCompute((kwargs["ret"][0], rt.SetTo, a << b))
            return kwargs["ret"][0]
        return a << b
    ret = kwargs["ret"][0] if "ret" in kwargs else ev.EUDVariable()
    if isinstance(a, ev.EUDVariable) and not isinstance(b, ev.EUDVariable):
        if b == 0:
            if a is ret:
                return a
            ev.SeqCompute([(ret, rt.SetTo, a)])
        elif 1 <= b <= 10:
            try:
                f = _fdict[b]
            except KeyError:
                set_ret = _selfadder.SetDest(0)

                @_EUDPredefineParam((EPD(set_ret) + 5, _selfadder))
                @EUDFunc
                def f(ret, adder):
                    for i in range(b):
                        ev.VProc(adder, [])
                    ev.VProc(
                        adder,
                        [
                            set_ret,
                            adder.SetModifier(rt.SetTo),
                        ],
                    )
                    rt.RawTrigger(
                        actions=[
                            adder.SetDest(EPD(adder.getValueAddr())),
                            adder.SetModifier(rt.Add),
                        ]
                    )

                _fdict[b] = f

            if ev.IsEUDVariable(ret):
                retarg = EPD(ret.getValueAddr())
            else:
                retarg = ret
            if "ret" in kwargs:
                del kwargs["ret"]
            f(retarg, a, **kwargs)
        else:
            if a is not ret:
                ev.SeqCompute([(ret, rt.SetTo, a)])
            ev.vbase.VariableBase.__ilshift__(ret, b)
        return ret

    return _f_bitlshift(a, b, **kwargs)


def f_bitrshift(a, b, **kwargs):
    """Calculate a >> b"""
    a, b = unProxy(a), unProxy(b)
    if not isinstance(a, ev.EUDVariable) and not isinstance(b, ev.EUDVariable):
        if "ret" in kwargs:
            SeqCompute((kwargs["ret"][0], rt.SetTo, a >> b))
            return kwargs["ret"][0]
        return a >> b
    ret = kwargs["ret"][0] if "ret" in kwargs else ev.EUDVariable()
    if isinstance(a, ev.EUDVariable) and not isinstance(b, ev.EUDVariable):
        if b == 0:
            if a is ret:
                return a
            ev.SeqCompute([(ret, rt.SetTo, a)])
        else:
            if a is not ret:
                ev.SeqCompute([(ret, rt.SetTo, a)])
            ev.vbase.VariableBase.__irshift__(ret, b)
        return ret

    return _f_bitrshift(a, b, **kwargs)



# Copyright 2014 by trgk.
# All rights reserved.
# This file is part of EUD python library (eudplib),
# and is released under "MIT License Agreement". Please see the LICENSE
# file that should have been included as part of this package.

from collections.abc import Iterator
from typing import ClassVar, Self

from ... import core as c
from ... import ctrlstru as cs
from ... import utils as ut
from ...localize import _
from ...maprw.injector.mainloop import _on_game_loop_start
from ...memio import (
    f_bread_cp,
    f_cunitepdread_epd,
    f_dwepdread_epd,
    f_epdcunitread_cp,
    f_setcurpl2cpcache,
)
from ...memio import modcurpl as cp
from ...scdata import CUnit
from ...utils import EPD
from .unlimiterflag import IsUnlimiterOn


def EUDLoopList(  # noqa: N802
    header_offset, break_offset=None
) -> Iterator[tuple[c.EUDVariable, c.EUDVariable]]:
    blockname = "listloop"
    ut.EUDCreateBlock(blockname, header_offset)

    ptr, epd = f_dwepdread_epd(ut.EPD(header_offset))

    if break_offset is not None:
        cs.EUDWhileNot()(ptr == break_offset)
    else:
        cs.EUDWhile()([ptr > 0, ptr <= 0x7FFFFFFF])

    yield ptr, epd
    cs.EUDSetContinuePoint()
    epd += 1
    f_dwepdread_epd(epd, ret=[ptr, epd])
    cs.EUDEndWhile()

    ut.ep_assert(
        ut.EUDPopBlock(blockname)[1] is header_offset, _("listloop mismatch")
    )


def EUDLoopUnit() -> Iterator[tuple[c.EUDVariable, c.EUDVariable]]:  # noqa: N802
    ut.EUDCreateBlock("unitloop", 0x628430)

    ptr, epd = f_cunitepdread_epd(EPD(0x628430))

    if cs.EUDWhileNot()(ptr == 0):
        yield ptr, epd
        cs.EUDSetContinuePoint()
        epd += 1
        f_cunitepdread_epd(epd, ret=[ptr, epd])
    cs.EUDEndWhile()

    ut.EUDPopBlock("unitloop")


class _UniqueIdentifier(c.EUDObject):
    _instances: ClassVar[list[Self]] = []

    def __new__(cls, *args, **kwargs) -> Self:
        return super().__new__(cls)

    def __init__(self) -> None:
        super().__init__()
        if not _UniqueIdentifier._instances:
            _on_game_loop_start(lambda: _UniqueIdentifier._init())
        _UniqueIdentifier._instances.append(self)

    def GetDataSize(self) -> int:  # noqa: N802
        return 4 + 336 * 1699

    def WritePayload(self, emitbuffer) -> None:  # noqa: N802
        emitbuffer.WriteDword(0)
        for _unit_data in range(1699):
            emitbuffer.WriteSpace(332)
            emitbuffer.WriteDword(0)

    @classmethod
    def _init(cls):
        first_empty_unit = EPD(0x628438)
        quot, rem = divmod(len(_UniqueIdentifier._instances), 32)
        endflags = c.Db((len(_UniqueIdentifier._instances) + 7) // 8)
        cs.DoActions(
            *[c.SetMemory(endflags + 4 * i, c.SetTo, 0) for i in range(quot)],
            c.SetMemoryX(endflags + 4 * quot, c.SetTo, 0, (1 << rem) - 1)
            if rem > 0
            else [],
            c.SetMemory(0x6509B0, c.SetTo, first_empty_unit),
        )
        epd = c.EUDXVariable(EPD(0x6509B0), c.Add, 0)
        f_epdcunitread_cp(0, ret=[epd])
        if cs.EUDWhileNot()(epd == 0):
            c.VProc(epd, c.SetMemory(0x6509B0, c.SetTo, 0xA5 // 4))
            uniq = f_bread_cp(0, 1)
            for i, unique_identifier in enumerate(_UniqueIdentifier._instances):
                q, r = divmod(i, 32)
                if cs.EUDIfNot()(c.MemoryX(endflags + 4 * q, c.AtLeast, 1, 1 << r)):
                    check_unique = c.DeathsX(cp.CP, c.Exactly, 0, 0, 0xFF)
                    c.VProc(
                        [epd, uniq],
                        [
                            c.SetMemory(
                                0x6509B0,
                                c.SetTo,
                                EPD(unique_identifier) - EPD(0x59CCA8),
                            ),
                            uniq.SetDest(EPD(check_unique + 8)),
                        ],
                    )
                    if cs.EUDIf()(check_unique):
                        cs.DoActions(
                            c.SetMemoryX(endflags + 4 * q, c.SetTo, 1 << r, 1 << r)
                        )
                    if cs.EUDElse()():
                        c.VProc(uniq, uniq.SetDest(cp.CP))
                    cs.EUDEndIf()
                cs.EUDEndIf()
            cs.EUDBreakIf(
                [
                    *[
                        c.Memory(endflags + 4 * i, c.Exactly, 0xFFFFFFFF)
                        for i in range(quot)
                    ],
                    c.MemoryX(
                        endflags + 4 * quot,
                        c.Exactly,
                        (1 << rem) - 1,
                        (1 << rem) - 1,
                    )
                    if rem > 0
                    else [],
                ]
            )
            cs.EUDSetContinuePoint()
            c.VProc(epd, c.SetMemory(0x6509B0, c.SetTo, 0x04 // 4))
            f_epdcunitread_cp(0, ret=[epd])
        cs.EUDEndWhile()
        f_setcurpl2cpcache()


def EUDLoopNewUnit(  # noqa: N802
    allowance: int = 2,
) -> Iterator[tuple[c.EUDVariable, c.EUDVariable]]:
    unique_identifier = _UniqueIdentifier()
    ut.ep_assert(isinstance(allowance, int))
    first_unit_ptr = EPD(0x628430)
    ut.EUDCreateBlock("newunitloop", "newlo")
    tos0 = c.EUDLightVariable(0)
    # tos0 << 0

    ptr, epd = f_cunitepdread_epd(first_unit_ptr)
    if cs.EUDWhileNot()(ptr == 0):
        c.VProc(
            epd,
            [
                c.SetMemory(0x6509B0, c.SetTo, 0xA5 // 4),
                epd.QueueAddTo(EPD(0x6509B0)),
            ],
        )
        uniq = f_bread_cp(0, 1)
        check_unique = c.DeathsX(cp.CP, c.Exactly, 0, 0, 0xFF)
        c.VProc(
            uniq,
            [
                c.SetMemory(
                    0x6509B0,
                    c.Add,
                    EPD(unique_identifier) - EPD(0x59CCA8) - 0xA5 // 4,
                ),
                uniq.SetDest(EPD(check_unique + 8)),
            ],
        )
        if cs.EUDIfNot()(check_unique):
            f_setcurpl2cpcache(uniq, uniq.SetDest(cp.CP))
            yield ptr, epd
        if cs.EUDElse()():
            tos0 += 1
            cs.EUDBreakIf(tos0.AtLeast(allowance))
        cs.EUDEndIf()
        cs.EUDSetContinuePoint()
        epd += 1
        f_cunitepdread_epd(epd, ret=[ptr, epd])
    cs.EUDEndWhile()

    f_setcurpl2cpcache([], tos0.SetNumber(0))
    ut.EUDPopBlock("newunitloop")


def EUDLoopUnit2() -> Iterator[tuple[c.EUDVariable, c.EUDVariable]]:  # noqa: N802
    """EUDLoopUnit보다 약간? 빠릅니다. 유닛 리스트를 따라가지 않고
    1700개 유닛을 도는 방식으로 작동합니다.
    """
    if IsUnlimiterOn():
        offset, is_dead = 0x4C // 4, c.MemoryXEPD(0, c.Exactly, 0, 0xFF00)
    else:
        offset, is_dead = 0x0C // 4, c.MemoryEPD(0, c.Exactly, 0)
    ptr = c.EUDVariable()
    epd = c.EUDVariable()
    set_ptr, set_epd = ptr.SetNumber(0), epd.SetNumber(0)

    cs.DoActions(
        ptr.SetNumber(0x59CCA8),
        c.SetMemory(set_ptr + 20, c.SetTo, 0x59CCA8),
        epd.SetNumber(EPD(0x59CCA8)),
        c.SetMemory(is_dead + 4, c.SetTo, EPD(0x59CCA8) + offset),
        c.SetMemory(set_epd + 20, c.SetTo, EPD(0x59CCA8)),
    )
    if cs.EUDWhileNot()(ptr >= 0x59CCA8 + 336 * 1699 + 1):
        cs.EUDContinueIf(is_dead)
        yield ptr, epd

        cs.EUDSetContinuePoint()
        c.RawTrigger(
            actions=[
                c.SetMemory(is_dead + 4, c.Add, 84),
                c.SetMemory(set_ptr + 20, c.Add, 336),
                set_ptr,
                c.SetMemory(set_epd + 20, c.Add, 84),
                set_epd,
            ],
        )
    cs.EUDEndWhile()


def EUDLoopPlayerUnit(player) -> Iterator[tuple[c.EUDVariable, c.EUDVariable]]:  # noqa: N802
    player = c.EncodePlayer(player)
    first_player_unit = 0x6283F8
    ut.EUDCreateBlock("playerunitloop", first_player_unit)
    ptr, epd = f_cunitepdread_epd(ut.EPD(first_player_unit) + player)

    if cs.EUDWhile()(ptr >= 1):
        next_unit, nextptr = c.Forward(), c.Forward()
        # /*0x06C*/ BW::CUnit*  nextPlayerUnit;
        # f_cunitepdread_epd(
        #     epd + 0x6C // 4,
        #     ret=[EPD(next_unit) + 5, EPD(next_unit) + 13]
        # )
        c.VProc(
            epd,
            [
                c.SetMemory(0x6509B0, c.SetTo, 0x6C // 4),
                epd.QueueAddTo(EPD(0x6509B0)),
                f_cunitepdread_epd._frets[0].SetDest(EPD(next_unit) + 5),
                f_cunitepdread_epd._frets[1].SetDest(EPD(next_unit) + 13),
                c.SetMemory(f_cunitepdread_epd._nptr, c.SetTo, nextptr),
            ],
        )
        c.SetNextTrigger(f_cunitepdread_epd._fstart)
        nextptr << c.NextTrigger()
        yield ptr, epd
        cs.EUDSetContinuePoint()
        cs.DoActions(next_unit << ptr.SetNumber(0), epd.SetNumber(0))
    cs.EUDEndWhile()

    ut.EUDPopBlock("playerunitloop")


def EUDLoopBullet() -> Iterator[tuple[c.EUDVariable, c.EUDVariable]]:  # noqa: N802
    yield from EUDLoopList(0x64DEC4)


def EUDLoopSprite() -> Iterator[tuple[c.EUDVariable, c.EUDVariable]]:  # noqa: N802
    y_epd = c.EUDVariable()
    y_epd << ut.EPD(0x629688)

    ut.EUDCreateBlock("spriteloop", "sprlo")

    if cs.EUDWhile()(y_epd < ut.EPD(0x629688) + 256):
        ptr, epd = f_dwepdread_epd(y_epd)
        if cs.EUDWhile()(ptr >= 1):
            yield ptr, epd
            cs.EUDSetContinuePoint()
            epd += 1
            f_dwepdread_epd(epd, ret=[ptr, epd])
        cs.EUDEndWhile()
        y_epd += 1
    cs.EUDEndWhile()

    ut.EUDPopBlock("spriteloop")


def EUDLoopNewCUnit(allowance: int = 2) -> Iterator[CUnit]:  # noqa: N802
    for ptr, epd in EUDLoopNewUnit(allowance):
        yield CUnit.cast(epd, ptr=ptr)


def EUDLoopCUnit() -> Iterator[CUnit]:  # noqa: N802
    """EUDLoopUnit보다 약간? 빠릅니다. 유닛 리스트를 따라가지 않고
    1700개 유닛을 도는 방식으로 작동합니다.
    """
    for ptr, epd in EUDLoopUnit2():
        yield CUnit.cast(epd, ptr=ptr)


def EUDLoopPlayerCUnit(player) -> Iterator[CUnit]:  # noqa: N802
    for ptr, epd in EUDLoopPlayerUnit(player):
        yield CUnit.cast(epd, ptr=ptr)
